"""Monaco Editors package.

This package provides integration and configuration for Monaco Editor components,
including language servers and editor models.
"""

from .base import monaco_editor
from .lifespan_tasks import start_terraform_ls
from .models import Command, LanguageClientConfig, LanguageServerUrl, TextModel

__all__ = (
    "Command",
    "LanguageClientConfig",
    "LanguageServerUrl",
    "TextModel",
    "monaco_editor",
    "start_terraform_ls",
)



"""Constants and configuration snippets for Monaco editor integration."""

from types import SimpleNamespace
from typing import Final


class CodeLensProviders(SimpleNamespace):
    """Constants containing JavaScript code snippets for registering CodeLens providers in Monaco editor."""

    TERRAFORM_RESOURCE_DOCS: Final = """const commandId = editor.addCommand(
        0,
        function (...args) {
            let [_, url] = args;
            window.open(url, '_blank').focus();
        },
    );
    vscode.languages.registerCodeLensProvider("terraform", {
        provideCodeLenses: async (document, _token) => {
            const providers = await getProviders();
            if (!providers) {
                return undefined;
            }
            const lines = [...Array(document.lineCount).keys()];
            const lenses = lines.filter(
                line => document.lineAt(line).text.startsWith("resource", 0)
            ).map(line => {
                let lineObj = document.lineAt(line);
                for (let key in providers) {
                    if (lineObj.text.includes(`${key}_`, 0)) {
                        const [_, resource, __] = lineObj.text.split(" ");
                        let urlResource = resource.replace(`${key}_`, "").replace(/"/g, '')
                        let url = `https://registry.terraform.io/providers/${providers[key]["owner"]}/${key}/${providers[key]["version"]}/docs/resources/${urlResource}`
                        let docString = urlResource.split("_").join(" ").toTitleCase();
                        return {
                            range: lineObj.range,
                            command: {
                                command: commandId,
                                title: docString,
                                arguments: [url],
                                tooltip: "Click to open documentation in new tab"
                            },
                        }
                    }
                }
            });
            return lenses;
        },
        resolveCodeLens: function (model, codeLens, token) {
            console.log(model, codeLens, token)
            return codeLens;
        },
    });
    """


class FunctionConstants(SimpleNamespace):
    """Constants containing JavaScript functions and configuration snippets for Monaco editor integration."""

    CONTAINER_REF: Final = "const [container, setContainer] = useState(null);"
    WORKSPACE: Final = "const workspace = `${{{workspace_folder}}}`;"
    CODE_VALUE: Final = "const codeValue = `${{{value}}}`;"
    GET_PROVIDERS: Final = """const getProviders = async () => {
        const client = wrapper.getLanguageClient("terraform");
        const _providers = await client.sendRequest("workspace/executeCommand", {
            "command": 'terraform-ls.module.providers',
            "arguments": [`uri=${vscode.Uri.parse(workspace)}`],
        });
        let providerMap = {}
        for (let key in _providers["installed_providers"]) {
            const [registry, owner, provider] = key.split("/")
            if (registry == "registry.terraform.io") {
                providerMap[provider] = {
                    owner: owner,
                    registry: registry,
                    version: _providers["installed_providers"][key]
                }
            }
        }
        return providerMap || undefined
    };
    """
    USER_CONFIG: Final = """const userConfiguration = {{
        'workbench.colorTheme': {theme},
        'editor.guides.bracketPairsHorizontal': 'active',
        'editor.experimental.asyncTokenization': true,
    }};
    """
    REGISTER_COMMANDS: Final = """
    const registerCommand = async ({{language, type, name, method, params, restart_client}}) => {{
        const registered = await vscode.commands.getCommands(true);
        if (!registered.includes(name)) {{
            if (type === "request") {{
                vscode.commands.registerCommand(name, async () => {{
                    {on_command}
                    const languageClient = wrapper.getLanguageClient(language);
                    await languageClient.sendRequest(method, params);
                    if (restart_client) {{
                        {on_restart}
                        await wrapper.getLanguageClientWrapper(language).restartLanguageClient();
                    }};
                    {on_command_complete}
                }});
            }} else {{
                vscode.commands.registerCommand(name, async () => {{
                    {on_command}
                    const languageClient = wrapper.getLanguageClient(language);
                    await languageClient.sendNotification(method, params);
                    if (restart_client) {{
                        {on_restart}
                        await wrapper.getLanguageClientWrapper(language).restartLanguageClient();
                    }};
                    {on_command_complete}
                }});
            }}
        }}
    }};
    """


class UseEffects(SimpleNamespace):
    """Constants containing JavaScript useEffect hooks."""

    UPDATE_USER_CONFIG: Final = """useEffect(() => {
        if (wrapper.isStarted()) {
            (async() => {
                await updateUserConfiguration(JSON.stringify(userConfiguration));
            })();
        }
    }, [wrapper, userConfiguration]);
    """
    INIT_WRAPPER: Final = """useEffect(() => {{
        if (container && !wrapper.isStarted()) {{
            (async () => {{
                wrapperConfig.htmlContainer = container;
                await wrapper.init(wrapperConfig);
                {text_change_callback}
                await wrapper.start();
                const editor = wrapper.getEditor();
                {additional}
                setStarted(true);
            }})();
        }}
    }}, [wrapper, container, started]);
    """
    UPDATE_CODE: Final = """useEffect(() => {{
        (async () => {{
            await wrapper.updateCodeResources({{
                modified: {{
                    text: codeValue,
                    uri: `${{workspace}}/${{{filename}}}`,
                }}
            }});
        }})();
    }}, [codeValue]);
    """


class WrapperConfig(SimpleNamespace):
    """Configuration constants for initializing and customizing the Monaco editor wrapper."""

    BASE: Final = """const wrapperConfig = {{
        $type: 'extended',
        logLevel: LogLevel.{loglevel},
        vscodeApiConfig: {vscode_api_config},
        editorAppConfig: {editor_app_config},
        languageClientConfigs: {language_client_configs}
    }};
    """
    VSCODE_API_CONFIG: Final = """{
        viewsConfig: {
            viewServiceType: 'EditorService'
        },
        serviceOverrides: {
            ...getKeybindingsServiceOverride(),
            ...getExtensionServiceOverride(),
        },
        userConfiguration: {
            json: JSON.stringify(userConfiguration)
        },
    }"""
    EDITOR_APP_CONFIG: Final = """{{
        monacoWorkerFactory: configureDefaultWorkerFactory,
        codeResources: {{
            modified: {{
                text: codeValue,
                uri: `${{workspace}}/${{{filename}}}`,
            }}
        }}
    }}"""



"""Async context managers for managing Terraform Language Server and LSP WebSocket proxy lifecycle."""

from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from typing import Any

from reflex.utils.processes import new_process

from .terraform import download_lsp_ws_proxy, download_terraform_ls, get_bin_dir


@asynccontextmanager
async def start_terraform_ls(port: int = 9999) -> AsyncGenerator[None, Any, None]:
    """Starts the Terraform Language Server and LSP WebSocket proxy as an async context manager.

    Args:
        port (int, optional): Port to bind the LSP WebSocket proxy. Defaults to 9999.

    Yields:
        None: Yields control while the server is running.
    """
    download_terraform_ls()
    download_lsp_ws_proxy()
    bin_dir = get_bin_dir()
    cmd = f"./lsp-ws-proxy -l 0.0.0.0:{port} -s -- ./terraform-ls serve"
    proc = new_process(cmd, show_logs=True, shell=True, cwd=bin_dir)  # noqa: S604
    yield
    proc.terminate()
    proc.wait()


__all__ = ("start_terraform_ls",)



"""Plugins for integrating Monaco Editor with VSCode extensions in Reflex.

This module provides configuration classes and Vite plugin definitions
to enable Monaco Editor with VSCode language extensions and services.
"""

from vite_config_plugin import RawJS, ViteConfig


class MonacoEditorsReflexConfig:
    """A Reflex plugin for integrating Monaco Editor with VSCode extensions.

    This plugin provides the necessary Vite configuration and dependencies
    to use Monaco Editor with various VSCode language extensions and services
    in a Reflex application.
    """

    @classmethod
    def get_dependencies(cls) -> list[str]:
        """Return a list of dependencies required for the Monaco Editor Reflex plugin."""
        return [
            "@codingame/esbuild-import-meta-url-plugin@1.0.3",
            "@codingame/monaco-vscode-rollup-vsix-plugin@20.2.1",
        ]

    @classmethod
    def get_imports(cls) -> list[str]:
        """Return a list of custom import statements for the Monaco Editor Reflex plugin."""
        return [
            "import importMetaUrlPlugin from '@codingame/esbuild-import-meta-url-plugin';",
            "import vsixPlugin from '@codingame/monaco-vscode-rollup-vsix-plugin';",
        ]

    @classmethod
    def get_vite_config(cls) -> ViteConfig:
        """Return the Vite configuration for integrating Monaco Editor with VSCode extensions.

        Returns:
            The configuration dictionary for Vite.
        """
        return {
            "worker": {"format": "es"},
            "assetsInclude": ["**/*.vsix"],
            "optimizeDeps": {
                "include": [
                    "@codingame/monaco-vscode-api",
                    "@codingame/monaco-vscode-configuration-service-override",
                    "@codingame/monaco-vscode-editor-api",
                    "@codingame/monaco-vscode-editor-service-override",
                    "@codingame/monaco-vscode-extension-api",
                    "@codingame/monaco-vscode-extensions-service-override",
                    "@codingame/monaco-vscode-java-default-extension",
                    "@codingame/monaco-vscode-javascript-default-extension",
                    "@codingame/monaco-vscode-json-default-extension",
                    "@codingame/monaco-vscode-languages-service-override",
                    "@codingame/monaco-vscode-localization-service-override",
                    "@codingame/monaco-vscode-model-service-override",
                    "@codingame/monaco-vscode-monarch-service-override",
                    "@codingame/monaco-vscode-standalone-languages",
                    "@codingame/monaco-vscode-textmate-service-override",
                    "@codingame/monaco-vscode-theme-defaults-default-extension",
                    "@codingame/monaco-vscode-theme-service-override",
                    "@codingame/monaco-vscode-views-service-override",
                    "@codingame/monaco-vscode-workbench-service-override",
                    "vscode/localExtensionHost",
                    "vscode-textmate",
                    "vscode-oniguruma",
                    "@vscode/vscode-languagedetection",
                ],
                "esbuildOptions": {
                    "plugins": [RawJS("importMetaUrlPlugin")],
                },
            },
            "plugins": [RawJS("vsixPlugin()")],
        }



"""Base module for Monaco editor integration with Reflex."""

from typing import Literal

import reflex as rx

from monaco_editors import constants

from .models import Command, LanguageClientConfig, LanguageServerUrl, TextModel


def generate_start_options(config: LanguageClientConfig) -> str:
    """Generates the JS `onCall` start options for the language client, if commands are configured for the editor.

    Args:
        config (LanguageClientConfig): The language client config object.

    Returns:
        The configured `startOptions` for the language client as a JS object string.
    """
    if config.register_commands:
        commands = f"""Object.entries({config.register_commands}).map(async ([name, params]) => {{
            params["name"] = name
            params["language"] = '{config.language_id}'
            await registerCommand(params);
        }})"""
        return f"""startOptions: {{
            onCall: async () => {{
                {commands}
            }}
        }}
        """
    return ""


def configure_language_clients(language_clients: list[LanguageClientConfig]) -> str:
    """Configures all language clients as JS objects.

    Args:
        language_clients (list[LanguageClientConfig]): The list of lanaguge clients to configure for the editor.

    Returns:
        The language client configs JS object string.
    """
    if language_clients:
        config_string = ""
        for config in language_clients:
            config_string += f"""{config.language_id}: {{
                name: "{config.language_id} client",
                connection: {{
                    options: {{
                        $type: "WebSocketUrl",
                        url: "{config.url.formatted}",
                        {generate_start_options(config=config)}
                    }}
                }},
                clientOptions: {{
                    documentSelector: {config.document_selector or [config.language_id]},
                    workspaceFolder: {{
                        index: 0,
                        name: "workspace",
                        uri: vscode.Uri.parse(`${{workspace}}`)
                    }},
                    initializationOptions: {config.initialization_options or "{}"}
                }}
            }},
            """
        return f"""{{
            configs: {{
                {config_string}
            }}
        }}"""
    return "undefined"


class MonacoEditorReactComp(rx.Component):
    """Monaco Editor with configurable langauage clients."""

    library = "monaco-languageclient@9.11.0"
    lib_dependencies = [  # noqa: RUF012
        "monaco-editor-wrapper@6.12.0",
        "@codingame/monaco-vscode-api@20.2.1",
        "@codingame/monaco-vscode-all-language-default-extensions@20.2.1",
    ]

    def add_imports(self) -> dict:
        """Add imports."""
        return {
            "@codingame/monaco-vscode-api": rx.ImportVar("LogLevel", is_default=False, install=False),
            "monaco-editor-wrapper": [
                rx.ImportVar("MonacoEditorLanguageClientWrapper", is_default=False, install=False),
                rx.ImportVar(
                    "configureDefaultWorkerFactory",
                    is_default=False,
                    install=False,
                    package_path="/workers/workerLoaders",
                ),
            ],
            "vscode": rx.ImportVar("*", alias="vscode", is_default=True, install=False),
            "@codingame/monaco-vscode-keybindings-service-override": rx.ImportVar(
                "getKeybindingsServiceOverride", is_default=True, install=False
            ),
            "@codingame/monaco-vscode-extensions-service-override": rx.ImportVar(
                "getExtensionServiceOverride", is_default=True, install=False
            ),
            "@codingame/monaco-vscode-configuration-service-override": rx.ImportVar(
                "updateUserConfiguration", is_default=False, install=False
            ),
            "": [
                "@" + rx.asset("hashicorp-terraform.vsix", shared=True),
            ],
        }

    def add_hooks(self) -> list:
        """Add component function hooks."""
        # Internal Hooks - Does not have SELF access, so can only be static strings.
        internal = [
            rx.vars.base.Var(
                constants.FunctionConstants.CONTAINER_REF,
                _var_data=rx.vars.base.VarData(
                    imports={"react": ["useState"]}, position=rx.constants.Hooks.HookPosition.INTERNAL
                ),
            ),
            rx.vars.base.Var(
                "const [started, setStarted] = useState(false);",
                _var_data=rx.vars.base.VarData(
                    imports={"react": ["useState"]}, position=rx.constants.Hooks.HookPosition.INTERNAL
                ),
            ),
        ]

        # Pre-Trigger Hooks - mostly function `const` definitions.
        on_command = (
            f"{rx.vars.LiteralVar.create(self.event_triggers['on_command'])._js_expr}(name);"  # noqa: SLF001
            if self.event_triggers.get("on_command")
            else ""
        )
        on_command_complete = (
            f"{rx.vars.LiteralVar.create(self.event_triggers['on_command_complete'])._js_expr}(name);"  # noqa: SLF001
            if self.event_triggers.get("on_command_complete")
            else ""
        )
        on_restart = (
            f"{rx.vars.LiteralVar.create(self.event_triggers['on_restart'])._js_expr}(language);"  # noqa: SLF001
            if self.event_triggers.get("on_restart")
            else ""
        )

        if isinstance(self.workspace_folder, type(None)):  # noqa: FURB168
            self.workspace_folder = "/workspace"

        pre_triggers = [
            rx.vars.base.Var(
                pre_trigger,
                _var_data=rx.vars.base.VarData(position=rx.constants.Hooks.HookPosition.PRE_TRIGGER),
            )
            for pre_trigger in (
                constants.FunctionConstants.WORKSPACE.format(
                    workspace_folder=(
                        rx.Var.create(self.workspace_folder)
                        if isinstance(self.workspace_folder, str)
                        else self.workspace_folder
                    )
                ),
                constants.FunctionConstants.CODE_VALUE.format(
                    value=rx.Var.create(self.value) if isinstance(self.value, str) else self.value,
                ),
                constants.FunctionConstants.USER_CONFIG.format(theme=self.theme),
                constants.FunctionConstants.GET_PROVIDERS,
                constants.FunctionConstants.REGISTER_COMMANDS.format(
                    on_command=on_command, on_command_complete=on_command_complete, on_restart=on_restart
                ),
            )
        ]

        # Post-Trigger hooks - mostly `useEffect` functions to dynamically configure editor

        text_change_callback = (
            f"wrapper.registerTextChangedCallback({rx.vars.LiteralVar.create(self.event_triggers['on_change'])._js_expr})"  # noqa: SLF001
            if self.event_triggers.get("on_change")
            else ""
        )
        additional = f"{constants.CodeLensProviders.TERRAFORM_RESOURCE_DOCS}"

        post_triggers = [
            rx.vars.base.Var(
                post_trigger,
                _var_data=rx.vars.base.VarData(
                    imports={"react": ["useEffect"]}, position=rx.constants.Hooks.HookPosition.POST_TRIGGER
                ),
            )
            for post_trigger in (
                constants.WrapperConfig.BASE.format(
                    loglevel=self.loglevel,
                    vscode_api_config=constants.WrapperConfig.VSCODE_API_CONFIG,
                    editor_app_config=constants.WrapperConfig.EDITOR_APP_CONFIG.format(
                        filename=rx.Var.create(self.filename)
                    ),
                    language_client_configs=configure_language_clients(self.language_clients),
                ),
                constants.UseEffects.UPDATE_USER_CONFIG,
                constants.UseEffects.INIT_WRAPPER.format(
                    text_change_callback=text_change_callback, additional=additional
                ),
                constants.UseEffects.UPDATE_CODE.format(filename=rx.Var.create(self.filename)),
            )
        ]

        return [*internal, *pre_triggers, *post_triggers]

    def add_custom_code(self) -> list:
        """Returns custom JavaScript code snippets required for the Monaco editor component."""
        return [
            # Wrapper must be created once in the file rather than inside the
            # component function or the universe will explode.
            "const wrapper = new MonacoEditorLanguageClientWrapper();",
            # Adds a file-level String prototype function to title-case strings
            """String.prototype.toTitleCase = function () {
                return this.replace(
                    /\w\S*/g,
                    function(txt){return txt.charAt(0).toUpperCase() + txt.substr(1).toLowerCase();}
                );
            };
            """,  # noqa: W605
        ]

    def render(self) -> dict:
        """Render the Monaco editor component as a div with appropriate props.

        Returns:
            dict: A dictionary containing the component name and filtered props.
        """
        self.ref = self._ref
        rendered = super().render()
        forbidden_props = (
            "filename",
            "theme",
            "value",
            "workspaceFolder",
            "languageClients",
            "loglevel",
            "onChange",
            "onCommand",
            "onRestart",
            "onCommandComplete",
        )
        props = [prop for prop in rendered["props"] if not any(prop.startswith(i) for i in forbidden_props)]
        return {"name": '"div"', "props": props}

    # Reference used to pass to the monaco wrapper.
    _ref = rx.Var("setContainer")

    ###### Required Attributes ######
    # The name of the 'file' to open in the editor.
    filename: str | rx.Var[str]

    ###### Optional Attributes ######
    # Theme configured based on Reflex color mode condition.
    theme = rx.color_mode_cond(dark="Default Dark Modern", light="Default Light Modern")
    # The editor code content of the 'file'
    value: str | rx.Var[str] = ""
    # The name of the worspace/folder/directory of the 'file'.
    workspace_folder: str | rx.Var[str] | None = None
    # Any configured language clients for the editor.
    language_clients: list[LanguageClientConfig] = []  # noqa: RUF012
    # The monaco editor log level (logs to browser console).
    loglevel: Literal["Off", "Trace", "Debug", "Info", "Warning", "Error"] = "Info"
    # The HTML class of the editor window.
    class_name: str = "w-full h-full"

    ###### Configurable Event Handlers ######
    # Fires on editor code content change. Returns the contant as a TextModel object.
    on_change: rx.EventHandler[rx.event.passthrough_event_spec(TextModel)]
    # Fires when an user-registered editor command executes. Returns the name of the registered command.
    on_command: rx.EventHandler[rx.event.passthrough_event_spec(str)]
    # Fires when a language client is restarted. Returns the name of the editor langauge client that restarts.
    on_restart: rx.EventHandler[rx.event.passthrough_event_spec(str)]
    # Fires when an user-registered editor command finishes. Returns the name of the registered command.
    on_command_complete: rx.EventHandler[rx.event.passthrough_event_spec(str)]


class Monaco(rx.ComponentNamespace):
    """Namespace for Monaco editor components and configuration."""

    __call__ = MonacoEditorReactComp.create
    language_client = LanguageClientConfig
    server_url = LanguageServerUrl
    command = Command


monaco_editor = Monaco()



"""Functions for downloading and managing Terraform language server and LSP WebSocket proxy binaries.

This module provides utilities to fetch, extract, and prepare the necessary binaries for Terraform language server
and LSP WebSocket proxy, supporting multiple platforms and architectures.
"""

import functools
import io
import pathlib
import platform
import tarfile
import zipfile
from typing import Final, Literal

from reflex.utils import console, path_ops
from reflex.utils.decorator import once
from reflex.utils.net import get
from reflex.utils.prerequisites import get_backend_dir

TERRAFORM_EXTENSION_VERSION: Final = "v2.34.5"
LSP_WS_PROXY_VERSION: Final = "v0.8.0"


def get_bin_dir() -> pathlib.Path:
    """Returns the path to the binary directory, creating it if it does not exist.

    Returns:
        pathlib.Path: Path to the binary directory.
    """
    bin_dir = get_backend_dir() / "bin"
    path_ops.mkdir(bin_dir)
    return bin_dir


@once
def get_vscode_extension_package_json() -> dict:
    """Fetches the package.json for the HashiCorp Terraform VSCode Extension.

    Returns:
        dict: The parsed JSON content of the package.json file.
    """
    console.debug(f"Getting Package JSON for HashiCorp Terraform VSCode Extenstion {TERRAFORM_EXTENSION_VERSION}")
    response = get(
        f"https://raw.githubusercontent.com/hashicorp/vscode-terraform/refs/tags/{TERRAFORM_EXTENSION_VERSION}/package.json"
    )
    return response.json()


@functools.lru_cache(maxsize=2)
def get_platform(purpose: Literal["terraform-ls", "lsp-ws-proxy"]) -> str:
    """Determines the platform string for downloading binaries based on the system and purpose.

    Args:
        purpose (Literal["terraform-ls", "lsp-ws-proxy"]): The intended binary ("terraform-ls" or "lsp-ws-proxy").

    Returns:
        str: The platform identifier string used in download URLs.
    """
    match platform.system():
        case "Darwin":
            return "darwin" if purpose == "terraform-ls" else "macos"
        case "Windows":
            return "windows"
        case _:
            return "linux"


def download_terraform_ls() -> None:
    """Downloads and extracts the Terraform language server binary if not already present.

    This function determines the correct version and platform, fetches the binary archive,
    and extracts the executable to the bin directory.
    """
    bin_dir = get_bin_dir()
    terraform_ls_bin = bin_dir / "terraform-ls"
    if not terraform_ls_bin.exists():
        package_json = get_vscode_extension_package_json()
        terraform_ls_version = package_json["langServer"]["version"]
        os = get_platform("terraform-ls")
        arch = platform.machine()
        console.debug(f"Downloading terraform-ls v{terraform_ls_version} for {os} {arch}")
        base_url = "https://releases.hashicorp.com/terraform-ls/0.36.5/terraform-ls"
        response = get(f"{base_url}_{terraform_ls_version}_{os}_{arch}.zip")
        terraform_ls_zip = io.BytesIO(response.content)
        with zipfile.ZipFile(terraform_ls_zip, "r") as archive:
            for file in archive.filelist:
                if file.filename == "terraform-ls":
                    archive.extract(member=file, path=bin_dir)
        terraform_ls_bin.chmod(0o755)


def download_lsp_ws_proxy() -> None:
    """Downloads and extracts the LSP WebSocket proxy binary if not already present.

    This function determines the correct platform, fetches the binary archive,
    and extracts the executable to the bin directory.
    """
    bin_dir = get_bin_dir()
    if not (bin_dir / "lsp-ws-proxy").exists():
        os = get_platform("lsp-ws-proxy")
        console.debug(f"Downloading lsp-ws-proxy v{LSP_WS_PROXY_VERSION} for {os}")
        base_url = "https://github.com/qualified/lsp-ws-proxy/releases/download"
        response = get(f"{base_url}/{LSP_WS_PROXY_VERSION}/lsp-ws-proxy_{os}.tar.gz", follow_redirects=True)
        lsp_ws_proxy_tar = io.BytesIO(response.content)
        with tarfile.open(fileobj=lsp_ws_proxy_tar, mode="r:gz") as tarball:
            for file in tarball.getmembers():
                if file.name == "lsp-ws-proxy":
                    tarball.extract(file, bin_dir)



"""Models and configuration classes for Monaco editor integration with language servers."""

from typing import Annotated, Any, Literal, TypedDict

import reflex as rx
from pydantic import BaseModel, Field, PlainValidator


class Command(BaseModel):
    """Language Servers Commands.

    Used to register Monaco editor commands with the language client.

    Params:
        type (Literal["notification", "request"]): The type of LSP command to execute.
        method (str): The LSP method to send (based on the specific language server).
        params (dict[str, Any]): The parameters to send.
        restart_client: (bool): If the editor's language client should be restarted after command completion.
    """

    type: Literal["notification", "request"]
    method: str
    params: dict[str, Any]
    restart_client: bool = False


class LanguageServerUrl(BaseModel):
    """The language server URL.

    Params:
        host (str): The hostname with a language server websocket.
        port (int): The host's websocket port number.
        secured (bool): Whether the connection is secure (use `wss` instead of `ws`)
        path (str): Additional path URI for the langauge server.
    """

    host: str
    port: int
    secured: Annotated[bool, Field(default=True)]
    path: Annotated[str, Field(default="")]

    @property
    def formatted(self) -> str:
        """Return the formatted language server URL as a string."""
        schema = "wss" if self.secured else "ws"
        path = self.path
        if path and not path.startswith("/"):
            path = f"/{self.path}"
        return f"{schema}://{self.host}:{self.port}{path}"


def _var_validator(value: rx.Var) -> str:
    if not isinstance(value, rx.Var):
        msg = f"Value {value} must be passed as an rx.Var"
        raise TypeError(msg)
    return str(value)


class LanguageClientConfig(BaseModel):
    """The Monaco Language Client Config.

    Params:
        language_id (str): The language ID registered with the editor.
        url (LanguageServerUrl): The langauge server URL object.
        register_commands (dict[str, Command]): The mapping of command names and configs to register wuth the editor.
        document_selector (list[str | dict[str, str]] | None): The optional document selector for the LSP.
        initialization_options (dict): The language-specific LSP opts to provide the language server upon connection.
    """

    language_id: str
    url: LanguageServerUrl
    register_commands: Annotated[str, PlainValidator(_var_validator), Field(default="")]
    document_selector: Annotated[list[str | dict[str, str]] | None, Field(default=None)]
    initialization_options: Annotated[str, PlainValidator(_var_validator), Field(default="")]


class TextModel(TypedDict):
    """The response model sent by the editor's `onChange`.

    For standard (non-diff) editors, the updated code will be in the `modified` attribute.
    """

    modified: str
    original: str


__all__ = (
    "Command",
    "LanguageClientConfig",
    "LanguageServerUrl",
    "TextModel",
)
